@@ -17,9 +17,12 @@
 #include "SymbolTable.h"
 #include "SyntheticSections.h"
 #include "WriterUtils.h"
+#include "lld/Common/Arrays.h"
 #include "lld/Common/CommonLinkerContext.h"
 #include "lld/Common/Strings.h"
+#include "llvm/ADT/ArrayRef.h"
 #include "llvm/ADT/DenseSet.h"
+#include "llvm/ADT/MapVector.h"
 #include "llvm/ADT/SmallSet.h"
 #include "llvm/ADT/SmallVector.h"
 #include "llvm/ADT/StringMap.h"
@@ -30,6 +33,9 @@
 #include "llvm/Support/FormatVariadic.h"
 #include "llvm/Support/LEB128.h"
 #include "llvm/Support/Parallel.h"
+#include "llvm/Support/RandomNumberGenerator.h"
+#include "llvm/Support/SHA1.h"
+#include "llvm/Support/xxhash.h"
 
 #include <cstdarg>
 #include <map>
@@ -40,8 +46,7 @@
 using namespace llvm;
 using namespace llvm::wasm;
 
-namespace lld {
-namespace wasm {
+namespace lld::wasm {
 static constexpr int stackAlignment = 16;
 static constexpr int heapAlignment = 16;
 
@@ -63,6 +68,7 @@ private:
   void createStartFunction();
   void createApplyDataRelocationsFunction();
   void createApplyGlobalRelocationsFunction();
+  void createApplyTLSRelocationsFunction();
   void createApplyGlobalTLSRelocationsFunction();
   void createCallCtorsFunction();
   void createInitTLSFunction();
@@ -102,11 +108,12 @@ private:
 
   void writeHeader();
   void writeSections();
+  void writeBuildId();
 
   uint64_t fileSize = 0;
 
   std::vector<WasmInitEntry> initFunctions;
-  llvm::StringMap<std::vector<InputChunk *>> customSectionMapping;
+  llvm::MapVector<StringRef, std::vector<InputChunk *>> customSectionMapping;
 
   // Stable storage for command export wrapper function name strings.
   std::list<std::string> commandExportWrapperNames;
@@ -135,7 +142,7 @@ void Writer::calculateCustomSections() {
       // These custom sections are known the linker and synthesized rather than
       // blindly copied.
       if (name == "linking" || name == "name" || name == "producers" ||
-          name == "target_features" || name.startswith("reloc."))
+          name == "target_features" || name.starts_with("reloc."))
         continue;
       // These custom sections are generated by `clang -fembed-bitcode`.
       // These are used by the rust toolchain to ship LTO data along with
@@ -144,7 +151,7 @@ void Writer::calculateCustomSections() {
       if (name == ".llvmbc" || name == ".llvmcmd")
         continue;
       // Strip debug section in that option was specified.
-      if (stripDebug && name.startswith(".debug_"))
+      if (stripDebug && name.starts_with(".debug_"))
         continue;
       // Otherwise include custom sections by default and concatenate their
       // contents.
@@ -156,7 +163,7 @@ void Writer::calculateCustomSections() {
 void Writer::createCustomSections() {
   log("createCustomSections");
   for (auto &pair : customSectionMapping) {
-    StringRef name = pair.first();
+    StringRef name = pair.first;
     LLVM_DEBUG(dbgs() << "createCustomSection: " << name << "\n");
 
     OutputSection *sec = make<CustomSection>(std::string(name), pair.second);
@@ -218,6 +225,91 @@ void Writer::writeSections() {
   });
 }
 
+// Computes a hash value of Data using a given hash function.
+// In order to utilize multiple cores, we first split data into 1MB
+// chunks, compute a hash for each chunk, and then compute a hash value
+// of the hash values.
+
+static void
+computeHash(llvm::MutableArrayRef<uint8_t> hashBuf,
+            llvm::ArrayRef<uint8_t> data,
+            std::function<void(uint8_t *dest, ArrayRef<uint8_t> arr)> hashFn) {
+  std::vector<ArrayRef<uint8_t>> chunks = split(data, 1024 * 1024);
+  std::vector<uint8_t> hashes(chunks.size() * hashBuf.size());
+
+  // Compute hash values.
+  parallelFor(0, chunks.size(), [&](size_t i) {
+    hashFn(hashes.data() + i * hashBuf.size(), chunks[i]);
+  });
+
+  // Write to the final output buffer.
+  hashFn(hashBuf.data(), hashes);
+}
+
+static void makeUUID(unsigned version, llvm::ArrayRef<uint8_t> fileHash,
+                     llvm::MutableArrayRef<uint8_t> output) {
+  assert((version == 4 || version == 5) && "Unknown UUID version");
+  assert(output.size() == 16 && "Wrong size for UUID output");
+  if (version == 5) {
+    // Build a valid v5 UUID from a hardcoded (randomly-generated) namespace
+    // UUID, and the computed hash of the output.
+    std::array<uint8_t, 16> namespaceUUID{0xA1, 0xFA, 0x48, 0x2D, 0x0E, 0x22,
+                                          0x03, 0x8D, 0x33, 0x8B, 0x52, 0x1C,
+                                          0xD6, 0xD2, 0x12, 0xB2};
+    SHA1 sha;
+    sha.update(namespaceUUID);
+    sha.update(fileHash);
+    auto s = sha.final();
+    std::copy(s.data(), &s.data()[output.size()], output.data());
+  } else if (version == 4) {
+    if (auto ec = llvm::getRandomBytes(output.data(), output.size()))
+      error("entropy source failure: " + ec.message());
+  }
+  // Set the UUID version and variant fields.
+  // The version is the upper nibble of byte 6 (0b0101xxxx or 0b0100xxxx)
+  output[6] = (static_cast<uint8_t>(version) << 4) | (output[6] & 0xF);
+
+  // The variant is DCE 1.1/ISO 11578 (0b10xxxxxx)
+  output[8] &= 0xBF;
+  output[8] |= 0x80;
+}
+
+void Writer::writeBuildId() {
+  if (!out.buildIdSec->isNeeded())
+    return;
+  if (config->buildId == BuildIdKind::Hexstring) {
+    out.buildIdSec->writeBuildId(config->buildIdVector);
+    return;
+  }
+
+  // Compute a hash of all sections of the output file.
+  size_t hashSize = out.buildIdSec->hashSize;
+  std::vector<uint8_t> buildId(hashSize);
+  llvm::ArrayRef<uint8_t> buf{buffer->getBufferStart(), size_t(fileSize)};
+
+  switch (config->buildId) {
+  case BuildIdKind::Fast: {
+    std::vector<uint8_t> fileHash(8);
+    computeHash(fileHash, buf, [](uint8_t *dest, ArrayRef<uint8_t> arr) {
+      support::endian::write64le(dest, xxh3_64bits(arr));
+    });
+    makeUUID(5, fileHash, buildId);
+    break;
+  }
+  case BuildIdKind::Sha1:
+    computeHash(buildId, buf, [&](uint8_t *dest, ArrayRef<uint8_t> arr) {
+      memcpy(dest, SHA1::hash(arr).data(), hashSize);
+    });
+    break;
+  case BuildIdKind::Uuid:
+    makeUUID(4, {}, buildId);
+    break;
+  default:
+    llvm_unreachable("unknown BuildIdKind");
+  }
+  out.buildIdSec->writeBuildId(buildId);
+}
+
 static void setGlobalPtr(DefinedGlobal *g, uint64_t memoryPtr) {
   LLVM_DEBUG(dbgs() << "setGlobalPtr " << g->getName() << " -> " << memoryPtr << "\n");
   g->global->setPointerValue(memoryPtr);
@@ -347,7 +439,12 @@ void Writer::layoutMemory() {
     WasmSym::heapBase->setVA(memoryPtr);
   }
 
-  uint64_t maxMemorySetting = 1ULL << (config->is64.value_or(false) ? 48 : 32);
+  uint64_t maxMemorySetting = 1ULL << 32;
+  if (config->is64.value_or(false)) {
+    // TODO: Update once we decide on a reasonable limit here:
+    // https://github.com/WebAssembly/memory64/issues/33
+    maxMemorySetting = 1ULL << 34;
+  }
 
   if (config->initialMemory != 0) {
     if (config->initialMemory != alignTo(config->initialMemory, WasmPageSize))
@@ -450,6 +547,7 @@ void Writer::addSections() {
   addSection(out.nameSec);
   addSection(out.producersSec);
   addSection(out.targetFeaturesSec);
+  addSection(out.buildIdSec);
 }
 
 void Writer::finalizeSections() {
@@ -470,7 +568,7 @@ void Writer::populateTargetFeatures() {
   if (config->isPic) {
     // This should not be necessary because all PIC objects should
     // contain the mutable-globals feature.
-    // TODO(https://bugs.llvm.org/show_bug.cgi?id=52339)
+    // TODO (https://github.com/llvm/llvm-project/issues/51681)
     allowed.insert("mutable-globals");
   }
 
@@ -647,7 +745,7 @@ static bool shouldImport(Symbol *sym) {
   if (config->allowUndefinedSymbols.count(sym->getName()) != 0)
     return true;
 
-  return sym->importName.has_value();
+  return sym->isImported();
 }
 
 void Writer::calculateImports() {
@@ -895,13 +993,13 @@ static StringRef getOutputDataSegmentName(const InputChunk &seg) {
     return ".tdata";
   if (!config->mergeDataSegments)
     return seg.name;
-  if (seg.name.startswith(".text."))
+  if (seg.name.starts_with(".text."))
     return ".text";
-  if (seg.name.startswith(".data."))
+  if (seg.name.starts_with(".data."))
     return ".data";
-  if (seg.name.startswith(".bss."))
+  if (seg.name.starts_with(".bss."))
     return ".bss";
-  if (seg.name.startswith(".rodata."))
+  if (seg.name.starts_with(".rodata."))
     return ".rodata";
   return seg.name;
 }
@@ -911,7 +1009,7 @@ OutputSegment *Writer::createOutputSegment(StringRef name) {
   OutputSegment *s = make<OutputSegment>(name);
   if (config->sharedMemory)
     s->initFlags = WASM_DATA_SEGMENT_IS_PASSIVE;
-  if (!config->relocatable && name.startswith(".bss"))
+  if (!config->relocatable && name.starts_with(".bss"))
     s->isBss = true;
   segments.push_back(s);
   return s;
@@ -1043,14 +1141,31 @@ void Writer::createSyntheticInitFunctions() {
     }
   }
 
-  if (config->sharedMemory && out.globalSec->needsTLSRelocations()) {
-    WasmSym::applyGlobalTLSRelocs = symtab->addSyntheticFunction(
-        "__wasm_apply_global_tls_relocs", WASM_SYMBOL_VISIBILITY_HIDDEN,
-        make<SyntheticFunction>(nullSignature,
-                                "__wasm_apply_global_tls_relocs"));
-    WasmSym::applyGlobalTLSRelocs->markLive();
-    // TLS relocations depend on  the __tls_base symbols
-    WasmSym::tlsBase->markLive();
+  if (config->sharedMemory) {
+    if (out.globalSec->needsTLSRelocations()) {
+      WasmSym::applyGlobalTLSRelocs = symtab->addSyntheticFunction(
+          "__wasm_apply_global_tls_relocs", WASM_SYMBOL_VISIBILITY_HIDDEN,
+          make<SyntheticFunction>(nullSignature,
+                                  "__wasm_apply_global_tls_relocs"));
+      WasmSym::applyGlobalTLSRelocs->markLive();
+      // TLS relocations depend on  the __tls_base symbols
+      WasmSym::tlsBase->markLive();
+    }
+
+    auto hasTLSRelocs = [](const OutputSegment *segment) {
+      if (segment->isTLS())
+        for (const auto* is: segment->inputSegments)
+          if (is->getRelocations().size())
+            return true;
+      return false;
+    };
+    if (llvm::any_of(segments, hasTLSRelocs)) {
+      WasmSym::applyTLSRelocs = symtab->addSyntheticFunction(
+          "__wasm_apply_tls_relocs", WASM_SYMBOL_VISIBILITY_HIDDEN,
+          make<SyntheticFunction>(nullSignature,
+                                  "__wasm_apply_tls_relocs"));
+      WasmSym::applyTLSRelocs->markLive();
+    }
   }
 
   if (config->isPic && out.globalSec->needsRelocations()) {
@@ -1332,8 +1447,9 @@ void Writer::createApplyDataRelocationsFunction() {
     raw_string_ostream os(bodyContent);
     writeUleb128(os, 0, "num locals");
     for (const OutputSegment *seg : segments)
-      for (const InputChunk *inSeg : seg->inputSegments)
-        inSeg->generateRelocationCode(os);
+      if (!config->sharedMemory || !seg->isTLS())
+        for (const InputChunk *inSeg : seg->inputSegments)
+          inSeg->generateRelocationCode(os);
 
     writeU8(os, WASM_OPCODE_END, "END");
   }
@@ -1341,6 +1457,23 @@ void Writer::createApplyDataRelocationsFunction() {
   createFunction(WasmSym::applyDataRelocs, bodyContent);
 }
 
+void Writer::createApplyTLSRelocationsFunction() {
+  LLVM_DEBUG(dbgs() << "createApplyTLSRelocationsFunction\n");
+  std::string bodyContent;
+  {
+    raw_string_ostream os(bodyContent);
+    writeUleb128(os, 0, "num locals");
+    for (const OutputSegment *seg : segments)
+      if (seg->isTLS())
+        for (const InputChunk *inSeg : seg->inputSegments)
+          inSeg->generateRelocationCode(os);
+
+    writeU8(os, WASM_OPCODE_END, "END");
+  }
+
+  createFunction(WasmSym::applyTLSRelocs, bodyContent);
+}
+
 // Similar to createApplyDataRelocationsFunction but generates relocation code
 // for WebAssembly globals. Because these globals are not shared between threads
 // these relocation need to run on every thread.
@@ -1476,6 +1609,12 @@ void Writer::createInitTLSFunction() {
       writeU8(os, 0, "memory index immediate");
     }
 
+    if (WasmSym::applyTLSRelocs) {
+      writeU8(os, WASM_OPCODE_CALL, "CALL");
+      writeUleb128(os, WasmSym::applyTLSRelocs->getFunctionIndex(),
+                   "function index");
+    }
+
     if (WasmSym::applyGlobalTLSRelocs) {
       writeU8(os, WASM_OPCODE_CALL, "CALL");
       writeUleb128(os, WasmSym::applyGlobalTLSRelocs->getFunctionIndex(),
@@ -1530,6 +1669,7 @@ void Writer::createSyntheticSections() {
   out.elemSec = make<ElemSection>();
   out.producersSec = make<ProducersSection>();
   out.targetFeaturesSec = make<TargetFeaturesSection>();
+  out.buildIdSec = make<BuildIdSection>();
 }
 
 void Writer::createSyntheticSectionsPostLayout() {
@@ -1570,7 +1710,7 @@ void Writer::run() {
       sym->forceExport = true;
   }
 
-  // Delay reporting error about explicit exports until after
+  // Delay reporting errors about explicit exports until after
   // addStartStopSymbols which can create optional symbols.
   for (auto &name : config->requiredExports) {
     Symbol *sym = symtab->find(name);
@@ -1619,6 +1759,8 @@ void Writer::run() {
       createApplyDataRelocationsFunction();
     if (WasmSym::applyGlobalRelocs)
       createApplyGlobalRelocationsFunction();
+    if (WasmSym::applyTLSRelocs)
+      createApplyTLSRelocationsFunction();
     if (WasmSym::applyGlobalTLSRelocs)
       createApplyGlobalTLSRelocationsFunction();
     if (WasmSym::initMemory)
@@ -1689,6 +1831,7 @@ void Writer::run() {
 
   log("-- writeSections");
   writeSections();
+  writeBuildId();
   if (errorCount())
     return;
 
@@ -1722,5 +1865,4 @@ void Writer::createHeader() {
 
 void writeResult() { Writer().run(); }
 
-} // namespace wasm
-} // namespace lld
+} // namespace wasm::lld
